"""
    Log
"""
__author__ = ('Reza Zeiny <rezazeiny1998@gmail.com>',)

import json
import logging
from typing import List, Optional

import requests

from .color import get_color_text, print_color, Colors

try:
    # noinspection PyPep8Naming
    from persiantools.jdatetime import JalaliDateTime as datetime
except ImportError:
    print_color("Install persiantools library with pip install", color=Colors.RED_F)
    from datetime import datetime

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

is_set_logger: bool = False


class Logger:
    """
        Logger Class
    """
    logger = None
    app_name: str = None
    host: str = None
    extra_data: dict = None

    log_level: int = INFO
    log_format = '%(datetime)s %(level)s%(process_thread)s%(file_detail)s%(text)s'

    colorful: bool = True
    show_datetime: bool = True
    datetime_color: "Colors" = Colors.GREEN_F

    file_depth: int = 1
    show_file: bool = True
    show_line: bool = True
    show_func: bool = True
    file_color: "Colors" = Colors.MAGENTA_F

    show_process_name: bool = False
    show_process_id: bool = False
    show_thread_name: bool = False
    show_thread_id: bool = False
    process_thread_color: "Colors" = Colors.BLUE_F

    show_level: bool = True
    level_detail = {
        NOTSET: ("NOTSET   ", Colors.CYAN_F),
        CRITICAL: ("CRITICAL ", Colors.BRIGHT_RED_F),
        ERROR: ("ERROR    ", Colors.RED_F),
        WARNING: ("WARNING  ", Colors.BLUE_F),
        INFO: ("INFO     ", Colors.BRIGHT_CYAN_F),
        DEBUG: ("DEBUG    ", Colors.BRIGHT_MAGENTA_F),
    }

    log_server: str = None

    old_factory = None

    @classmethod
    def get_logger(cls, name: str, log_level: int = None):
        """
            For get logger
        """
        global is_set_logger
        cls.logger = logging.getLogger(name)
        if is_set_logger:
            return cls.logger
        if log_level is None:
            log_level = cls.log_level
        logging.basicConfig(format=cls.log_format, level=log_level)
        cls.old_factory = logging.getLogRecordFactory()
        logging.setLogRecordFactory(cls.record_factory)
        is_set_logger = True
        return cls.logger

    @classmethod
    def add_file_handler(cls, file_address: str, log_level=ERROR):
        """
            Add File Handler
        """
        if cls.logger is None:
            print_color("no logger found", color=Colors.RED_F)
            return
        handler = logging.FileHandler(file_address)
        handler.setLevel(log_level)
        handler.setFormatter(logging.Formatter(cls.log_format))
        cls.logger.addHandler(handler)

    @staticmethod
    def get_inner_detail(data_list: List[tuple], sep: str = ":"):
        """
            For remove repeat
        """
        detail = ""
        last_cond = False
        for i in range(len(data_list)):
            cond, data = data_list[i]
            if cond:
                if last_cond:
                    detail += sep
                detail += str(data)
            last_cond = cond
        return detail

    @staticmethod
    def get_outer_detail(data_list: [str], sep: str = " - "):
        """
            For remove repeat
        """
        detail = ""
        last_cond = False
        for i in range(len(data_list)):
            data = data_list[i]
            cond = data != ""
            if cond and last_cond:
                detail += sep
            detail += str(data)
            last_cond = cond
        if detail != "":
            return "[ " + detail + " ] "
        else:
            return detail

    @classmethod
    def get_color_text(cls, text, color: Optional[Colors]):
        """
            Get Color Text
        """
        if cls.colorful:
            return get_color_text(text, color=color)
        else:
            return str(text)

    @classmethod
    def get_datetime(cls):
        """
            Get Time
        """
        return cls.get_inner_detail([
            (cls.show_datetime, datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
        ])

    @classmethod
    def get_process_thread(cls, record):
        """
            Process and thread detail
        """
        process_detail = cls.get_inner_detail([
            (cls.show_process_name, record.processName),
            (cls.show_process_id, record.process)
        ])
        thread_detail = cls.get_inner_detail([
            (cls.show_thread_name, record.threadName),
            (cls.show_thread_id, record.thread)
        ])
        return cls.get_outer_detail([process_detail, thread_detail])

    @classmethod
    def get_file_detail(cls, record: logging.LogRecord):
        """
            Get File and line detail
        """
        file_detail = cls.get_inner_detail([
            (cls.show_file, "/".join(record.pathname.split("/")[-cls.file_depth:])),
            (cls.show_line, record.lineno),
        ])
        function_detail = cls.get_inner_detail([
            (cls.show_func, record.module + "()"),
        ])
        return cls.get_outer_detail([file_detail, function_detail])

    @classmethod
    def get_level(cls, record):
        """
            Get Level
        """
        return cls.get_inner_detail([
            (cls.show_level, cls.level_detail.get(record.levelno, ("",))[0])
        ])

    @classmethod
    def get_level_color(cls, record):
        """
            Get Level Color
        """
        return cls.level_detail.get(record.levelno, ("", Colors.YELLOW_B))[1]

    @classmethod
    def record_factory(cls, *args, **kwargs):
        """
        for change some value
        :param args:
        :param kwargs:
        :return:
        """
        new_args = list(args)
        message = str(new_args[4])
        message_args = new_args[5]
        extra_args = False
        if len(message_args) == 1 and "%s" not in message and "%d" not in message:
            new_args[4] += " %s"
            extra_args = True
        record = cls.old_factory(*tuple(new_args), **kwargs)
        record.datetime = cls.get_color_text(cls.get_datetime(), color=cls.datetime_color)
        record.process_thread = cls.get_color_text(cls.get_process_thread(record), color=cls.process_thread_color)
        record.file_detail = cls.get_color_text(cls.get_file_detail(record), color=cls.file_color)

        record.short_message = None
        record.full_message = None
        if type(record.msg) == dict:
            record.full_message = record.msg
        elif extra_args:
            record.short_message = args[4]
            data_args = new_args[5]
            if type(data_args[0]) == dict:
                record.full_message = data_args[0]
            else:
                record.full_message = str(data_args[0])
        else:
            # noinspection PyBroadException
            try:
                record.short_message = str(record.msg) % record.args
            except Exception:
                pass
        record.short_message = str(record.short_message)
        color = cls.get_level_color(record)
        record.level = cls.get_color_text(cls.get_level(record), color=color)
        record.text = cls.get_color_text(record.short_message, color=color)

        cls.send_to_log_server(record=record)
        return record

    @classmethod
    def get_send_data(cls, record: "logging.LogRecord"):
        """
            Get some log data
        """
        return {
            "level": record.levelno,
            "line": record.lineno,
            "file": record.filename,
            "level_name": record.levelname,
            "pathname": record.pathname,
            "thread": record.thread,
            "threadName": record.threadName,
            "process": record.process,
            "processName": record.processName,
            "function": record.module,
            "logger_name": record.name,
            "created": record.created,
            "msecs": record.msecs,
        }

    @classmethod
    def handle_log_server(cls, send_data):
        """
            For send data
        """
        if cls.log_server is None:
            return

        log_server = []
        if type(cls.log_server) == str:
            log_server = [cls.log_server]
        elif type(cls.log_server) in [list, tuple]:
            log_server = cls.log_server

        for server in log_server:
            try:
                headers = {"Content-Type": "application/json"}
                data = json.dumps(send_data)
                requests.post(server, data=data, headers=headers)
            except Exception as e:
                print_color(f"error in send data to log server {server}", e, color=Colors.RED_F)

    @classmethod
    def send_to_log_server(cls, record: "logging.LogRecord"):
        """
            Send Data to log server
        """
        if not cls.log_server:
            return
        short_message = record.__dict__.get("short_message")
        full_message = record.__dict__.get("full_message")
        send_data = {
            "short_message": short_message,
            "logger_data": cls.get_send_data(record)
        }
        if cls.app_name is not None:
            send_data["app_name"] = cls.app_name
        if cls.host is not None:
            send_data["host_name"] = cls.host
        if cls.extra_data is not None:
            send_data["extra_data"] = cls.extra_data
        urllib3_log = logging.getLogger("urllib3")
        urllib3_log.setLevel(logging.CRITICAL)
        urllib3_log.propagate = False

        if type(full_message) == dict:
            full_message = dict(full_message)
            for key, value in full_message.items():
                if str(key) not in ["logger_data", "extra_data", "short_message", "app_name", "host_name"]:
                    send_data[str(key)] = value

        cls.handle_log_server(send_data=send_data)



"""
Matrix Client
"""
import logging

from .api import MatrixHttpApi
from .errors import MatrixRequestError, MatrixUnexpectedResponse
from .room import Room

logger = logging.getLogger(__name__)


class MatrixClient(object):
    """
    The client API for Matrix. For the raw HTTP calls, see MatrixHttpApi.

    Args:
        token (Optional[str]): If you have an access token
            supply it here.
        valid_cert_check (bool): Check the home servers
            certificate on connections?
    Returns:
        `MatrixClient`
    Raises:
        `MatrixRequestError`, `ValueError`
    Examples:
        Create a new user and send a message::
            client = MatrixClient()
            token = client.login(username="foobar",
                password="monkey")
            room = client.create_room("my_room")
            room.send_image(file_like_object)
    """

    def __init__(self, base_url: str, *, token: str = None,
                 valid_cert_check: bool = True, sync_filter_limit: int = 20,
                 api_connection_timeout: int = 60, device_id: str = None):

        self.api = MatrixHttpApi(base_url=base_url, token=token,
                                 connection_timeout=api_connection_timeout, validate_cert=valid_cert_check)
        self.device_id = device_id

        self.sync_token = None
        self.sync_filter = '{ "room": { "timeline" : { "limit" : %i } } }' \
                           % sync_filter_limit
        self.token = token
        self.user_id = None
        self.home_server = None
        """ Time to wait before attempting a /sync request after failing."""
        self.rooms = {
            # room_id: Room
        }
        self.users = {
            # user_id: User
        }

    def login(self, username: str, password: str):
        """
        Login to the home server.

        Args:
            username (str): Account username
            password (str): Account password
        Returns:
            str: Access token
        Raises:
            MatrixRequestError
        """
        response = self.api.login(
            "m.login.password", user=username, password=password, device_id=self.device_id
        )
        self.user_id = response["user_id"]
        self.token = response["access_token"]
        self.home_server = response["home_server"]
        self.device_id = response["device_id"]
        self.api.token = self.token

        return self.token

    def logout(self):
        """
            Logout from the home server.
        """
        self.api.logout()

    # TODO: move room creation/joining to User class for future application service usage
    # NOTE: we may want to leave thin wrappers here for convenience
    def create_room(self, alias=None, is_public=False, invitees=None):
        """
        Create a new room on the home server.

        Args:
            alias (str): The canonical_alias of the room.
            is_public (bool):  The public/private visibility of the room.
            invitees (str[]): A set of user ids to invite into the room.

        Returns:
            Room

        Raises:
            MatrixRequestError
        """
        response = self.api.create_room(
            alias=alias, is_public=is_public, invitees=invitees)
        return self._make_room(response["room_id"])

    def join_room(self, room_id_or_alias):
        """
        Join a room.

        Args:
            room_id_or_alias (str): Room ID or an alias.

        Returns:
            Room

        Raises:
            MatrixRequestError
        """
        response = self.api.join_room(room_id_or_alias)
        room_id = (
            response["room_id"] if "room_id" in response else room_id_or_alias
        )
        return self._make_room(room_id)

    # TODO: move to User class. Consider creating lightweight Media class.
    def upload(self, content, content_type, filename=None):
        """ Upload content to the home server and receive a MXC url.

        Args:
            content (bytes): The data of the content.
            content_type (str): The mimetype of the content.
            filename (str): Optional. Filename of the content.

        Raises:
            MatrixUnexpectedResponse: If the home_server gave a strange response
            MatrixRequestError: If the upload failed for some reason.
        """
        try:
            response = self.api.media_upload(content, content_type, filename)
            if "content_uri" in response:
                return response["content_uri"]
            else:
                raise MatrixUnexpectedResponse(
                    "The upload was successful, but content_uri wasn't found."
                )
        except MatrixRequestError as e:
            raise MatrixRequestError(
                code=e.code,
                content="Upload failed: %s" % e
            )

    def _make_room(self, room_id):
        room = Room(self, room_id)
        self.rooms[room_id] = room
        return self.rooms[room_id]

    def sync(self, timeout_ms=30000):
        """
            For sync
        """
        response = self.api.sync(self.sync_token, timeout_ms=timeout_ms)
        self.sync_token = response["next_batch"]

        # for presence_update in response['presence']['events']:
        #     for callback in self.presence_listeners.values():
        #         callback(presence_update)

        # for room_id, invite_room in response['rooms']['invite'].items():
        #     for listener in self.invite_listeners:
        #         listener(room_id, invite_room['invite_state'])

        # for room_id, left_room in response['rooms']['leave'].items():
        #     for listener in self.left_listeners:
        #         listener(room_id, left_room)
        #     if room_id in self.rooms:
        #         del self.rooms[room_id]

        # if self._encryption and 'device_one_time_keys_count' in response:
        #     self.olm_device.update_one_time_key_counts(
        #         response['device_one_time_keys_count'])

        # for room_id, sync_room in response['rooms']['join'].items():
        #     if room_id not in self.rooms:
        #         self._make_room(room_id)
        #     room = self.rooms[room_id]
        #     # TODO: the rest of this for loop should be in room object method
        #     room.prev_batch = sync_room["timeline"]["prev_batch"]
        #
        #     for event in sync_room["state"]["events"]:
        #         event['room_id'] = room_id
        #         room._process_state_event(event)
        #
        #     for event in sync_room["timeline"]["events"]:
        #         event['room_id'] = room_id
        #         room._put_event(event)
        #
        #         # TODO: global listeners can still exist but work by each
        #         # room.listeners[uuid] having reference to global listener
        #
        #     for event in sync_room['ephemeral']['events']:
        #         event['room_id'] = room_id
        #         room._put_ephemeral_event(event)

    # TODO: move to Room class
    def remove_room_alias(self, room_alias):
        """Remove mapping of an alias

        Args:
            room_alias(str): The alias to be removed.

        Returns:
            bool: True if the alias is removed, False otherwise.
        """
        try:
            self.api.remove_room_alias(room_alias)
            return True
        except MatrixRequestError:
            return False



"""
    working with file easily
"""
__author__ = ('Reza Zeiny <rezazeiny1998@gmail.com>',)

import gzip
import json
import logging
import os
import pickle
from enum import Enum

logger = logging.getLogger(__name__)


class FileModes(Enum):
    """
        Type of Files
    """
    NORMAL = "normal"
    JSON = "json"
    GZIP = "gzip"
    OBJECT = "object"
    FILE = "file"
    APPEND = "append"


def read_file(address, file_mode=FileModes.NORMAL):
    """
        easy way to read file
    Args:
        address (str): file address
        file_mode (FileModes): mode of reading

    Returns:
        file data

    """
    if address is None:
        return None
    logger.debug(f"reading file. addr: {address} {file_mode}")
    mode = "r"
    if file_mode in [FileModes.OBJECT, FileModes.FILE]:
        mode = "rb"
    try:
        if file_mode == FileModes.GZIP:
            with gzip.open(address) as file:
                file_data = []
                for line in file.readlines():
                    try:
                        file_data.append(line.strip().decode("UTF-8"))
                    except UnicodeDecodeError:
                        pass
        else:
            with open(address, mode) as file:
                if file_mode == FileModes.JSON:
                    file_data = json.load(file)
                elif file_mode == FileModes.OBJECT:
                    file_data = pickle.load(file)
                elif file_mode == FileModes.FILE:
                    file_data = file.read()
                elif file_mode == FileModes.NORMAL:
                    file_data = [s.strip() for s in file.readlines()]

        logger.debug(f"reading file is complete. addr: {address} {file_mode}")
        return file_data
    except Exception as e:
        logger.error(f"reading file {address} failed e: {e}")
        return None


def write_file(address: str, data, file_mode=FileModes.NORMAL) -> bool:
    """
        Write or append easily to file
    Args:
        address (str): file address
        data: data
        file_mode (FileModes): file mode

    Returns:
        (bool) : job state
    """
    if address is None:
        return False

    if type(data) == str and file_mode in [FileModes.APPEND, FileModes.NORMAL]:
        data = [data]

    mode = "w"
    if file_mode == FileModes.APPEND:
        mode = "a+"
    elif file_mode in [FileModes.OBJECT, FileModes.FILE]:
        mode = "wb"
    logger.debug(f"mode: {file_mode}, addr: {address}")

    try:
        with open(address, mode) as file:
            if file_mode == FileModes.JSON:
                json.dump(data, file, indent=4, sort_keys=True, ensure_ascii=False)
            elif file_mode == FileModes.OBJECT:
                pickle.dump(data, file)
            elif file_mode == FileModes.FILE:
                file.write(data)
            else:
                for line in data:
                    file.write(str(line) + "\n")
        logger.debug(f"{file_mode} is complete. addr: {address}")

        return True
    except Exception as e:
        logger.error(f"{file_mode} fail. addr: {address}, e: {e}")
    return False


def remove_file_if_exists(path) -> bool:
    """
        Remove File if exists
        Args:
            path (str) : file path
        Returns:
            (bool) : remove state
    """
    if os.path.exists(path):
        os.remove(path)
        return True
    return False



"""
    Server Work
"""
__author__ = ('Reza Zeiny <zeiny.r@yaftar.ir>',)

import inspect
import os
from datetime import datetime
from typing import Tuple, Optional

import paramiko

from .color import Colors, print_color

try:
    import requests
except ImportError:
    requests = None

from sshtunnel import SSHTunnelForwarder

import logging
import time

logger = logging.getLogger(__name__)


class CustomSSHClient:
    """
        additional need for ssh
    """

    def __init__(self, ssh):
        self.ssh: "paramiko.SSHClient" = ssh

    def execute(self, command: str) -> Tuple[int,
                                             paramiko.channel.ChannelStdinFile,
                                             paramiko.channel.ChannelFile,
                                             paramiko.channel.ChannelStderrFile]:
        """
            Execute a command and get process id
        """
        command = 'echo $$; exec ' + command
        stdin, stdout, stderr = self.ssh.exec_command(command)
        pid = int(stdout.readline())
        return pid, stdin, stdout, stderr

    def kill(self, pid: int, signal: int = 15):
        """
            Kill a process
        """
        self.ssh.exec_command(f"kill -{signal} {pid}")


class CustomSFTPClient:
    """
        additional need for sftp
    """

    def __init__(self, sftp):
        self.sftp: "paramiko.SFTPClient" = sftp

        # def put_dir(self, local, remote):
        """ 
            Uploads the contents of the local directory to the remote path. The
            remote directory needs to exists. All subdirectories in local are 
            created under remote.
        """
        # for item in os.listdir(local):
        #     if os.path.isfile(os.path.join(local, item)):
        #         self.put(os.path.join(local, item), '%s/%s' % (remote, item))
        #     else:
        #         self.mkdir('%s/%s' % (remote, item), ignore_existing=True)
        #         self.put_dir(os.path.join(local, item), '%s/%s' % (remote, item))

    def get_dir(self, remote: str, local: str = "./"):
        """
            Uploads the contents of the remote directory to the local path. The
            local directory needs to exists. All subdirectories in remote are
            created under local.
        """
        if not os.path.exists(local):
            os.mkdir(local)
        for item in self.sftp.listdir(remote):
            remote_path = os.path.join(remote, item)
            local_path = os.path.join(local, item)
            remote_path_stat = str(self.sftp.stat(remote_path))
            if remote_path_stat[0] != 'd':
                self.sftp.get(remote_path, local_path)
            else:
                if not os.path.exists(local_path):
                    os.mkdir(local_path)
                self.get_dir(remote_path, local_path)

        # def mkdir(self, path, mode=511, ignore_existing=False):
        """ 
            Augments mkdir by adding an option to not fail if the folder exists  
        """
        # try:
        #     super(MySFTPClient, self).mkdir(path, mode)
        # except IOError:
        #     if ignore_existing:
        #         pass
        #     else:
        #         raise


class Server:
    """
        For connect to server ether before mirror
        SSH DOCS: http://docs.paramiko.org/en/stable/api/agent.html
        SFTP DOCS: http://docs.paramiko.org/en/stable/api/sftp.html
    """

    def __init__(self, *, name: str = "server", middle_host: str = None, middle_port: int = 22,
                 middle_username: str = "user",
                 middle_password: str = "1", host: str = "localhost", port: int = 22, username: str = "user",
                 password: str = "1", handle=None, next_host: str = None, next_port: int = None):
        self.__name = str(name)
        self.__middle_host = middle_host
        self.__middle_port = middle_port
        self.__middle_username = middle_username
        self.__middle_password = middle_password
        self.__host = host
        self.__port = port
        self.__username = username
        self.__password = password
        self.__ssh_host = host
        self.__ssh_port = port
        self.__ssh: Optional["paramiko.SSHClient"] = None
        self.__sftp: Optional["paramiko.SFTPClient"] = None
        self.__handle = handle
        self.__next_host = None
        self.__next_port = None
        self.__tunnel_host = next_host
        self.__tunnel_port = next_port

        self.start_dt = None
        self.end_dt = None

        self.is_connect = None
        self.return_value = None
        self.has_exception = None
        self.connection_time = None

        if callable(handle):
            print_color("handle in server class is deprecated. use handle in start", color=Colors.RED_F)

    def __str__(self):
        return f"'{self.__name}' {self.__username}@{self.__host}:{self.__port}"

    def __log(self, level, message, data=None):
        if data is None:
            data = {}
        log_data = {"server": {
            "username": self.__username,
            "ip": self.__host,
            "port": self.__port,
            **data
        }}
        if logger.isEnabledFor(level):
            # noinspection PyProtectedMember
            logger._log(level, f"{self}: {message}", (log_data,))

    @staticmethod
    def test_handle(sftp: "paramiko.SFTPClient"):
        return sftp.listdir()

    def get_handle_data(self, handle, **kwargs):
        """
            For get useful data for handle function
        """
        if not callable(handle):
            return None
        # noinspection PyUnresolvedReferences
        args_data = inspect.getfullargspec(handle)
        handle_inputs = args_data.args
        use_kwargs = args_data.varkw is not None
        kwargs = {
            "name": self.__name,
            "ssh": self.__ssh,
            "sftp": self.__sftp,
            "ssh_host": self.__ssh_host,
            "ssh_port": self.__ssh_port,
            "next_host": self.__next_host,
            "next_port": self.__next_port,
            **kwargs
        }
        send_data = {}
        for key, value in kwargs.items():
            if key in handle_inputs or use_kwargs:
                send_data[key] = value
        return send_data

    def _run(self, handle, retry_count, retry_delay, **kwargs):
        """
            Run all work
        """
        self.__log(logging.INFO, f"Connected {self.__ssh_host}:{self.__ssh_port}")
        self.is_connect = True
        for i in range(retry_count):
            try:
                self.has_exception = False
                handle_data = self.get_handle_data(handle=handle, **kwargs)
                self.return_value = handle(**handle_data)
                break
            except Exception as e:
                self.has_exception = True
                self.__log(logging.ERROR, f"error in calling function {handle.__name__} with error {e}")
            if i < (retry_count - 1):
                time.sleep(retry_delay)
        return self.return_value

    def start(self, handle=None, retry_count=1, retry_delay=0, **kwargs):
        """
            Start working
        """
        logging.getLogger("paramiko").setLevel(logging.WARNING)
        if handle is None:
            handle = self.__handle
        if handle is None:
            handle = self.test_handle
        if not callable(handle):
            self.__log(logging.ERROR, f"function {handle.__name__} is not callable")
            return
        self.is_connect = False
        self.start_dt = datetime.now()
        if self.__middle_host is None:
            self._set_connect(handle=handle, retry_count=retry_count, retry_delay=retry_delay, **kwargs)
        else:
            self._set_tunnel(handle=handle, retry_count=retry_count, retry_delay=retry_delay, **kwargs)
        self.end_dt = datetime.now()
        self.connection_time = (self.end_dt - self.start_dt).total_seconds()
        self.__log(logging.INFO, f"Run function {handle.__name__} is completed", data={
            "connection_time": self.connection_time,
            "is_connect": self.is_connect,
            "has_exception": self.has_exception
        })
        return self.return_value

    def _set_connect(self, handle, retry_count, retry_delay, **kwargs):
        """
            Start ssh connect
        """
        self.__log(logging.DEBUG, f"Start ssh {self.__ssh_host}:{self.__ssh_port}")
        try:
            if self.__tunnel_port is not None and self.__tunnel_host is not None:
                self.__log(logging.DEBUG, f"Start tunnel in for next host {self.__ssh_host}:{self.__ssh_port} => "
                                          f"{self.__tunnel_host}:{self.__tunnel_port}")
                with SSHTunnelForwarder(
                        ssh_address_or_host=(self.__ssh_host, self.__ssh_port),
                        ssh_username=self.__username,
                        ssh_password=self.__password,
                        remote_bind_address=(self.__tunnel_host, self.__tunnel_port),
                        set_keepalive=100, threaded=False) as tunnel:
                    self.__log(logging.DEBUG, f"Started tunneling success for next host "
                                              f"{tunnel.local_bind_host}:{tunnel.local_bind_port} => "
                                              f"{self.__tunnel_host}:{self.__tunnel_port}")
                    self.__next_host = tunnel.local_bind_host
                    self.__next_port = tunnel.local_bind_port
                    self._run(handle=handle, retry_count=retry_count, retry_delay=retry_delay, **kwargs)
                    self.__log(logging.DEBUG, f"Closing tunneling server for next host "
                                              f"{tunnel.local_bind_host}:{tunnel.local_bind_port} => "
                                              f"{self.__tunnel_host}:{self.__tunnel_port}")
            else:
                self.__ssh = paramiko.SSHClient()
                self.__ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self.__ssh.connect(self.__ssh_host, port=self.__ssh_port,
                                   username=self.__username,
                                   password=self.__password)
                self.__sftp: "paramiko.SFTPClient" = self.__ssh.open_sftp()
                self._run(handle=handle, retry_count=retry_count, retry_delay=retry_delay, **kwargs)
                try:
                    self.__ssh.close()
                    self.__log(logging.DEBUG, f"Closed {self.__ssh_host}:{self.__ssh_port}")
                except Exception as e:
                    self.__log(logging.ERROR, f"Error in closing ssh connection. error: {e}")

        except Exception as e:
            self.__log(logging.ERROR, f"Error in set connect. error: {e}")
        return self.return_value

    def _set_tunnel(self, handle, retry_count, retry_delay, **kwargs):
        """
            Start tunnel
        """
        self.__log(logging.DEBUG, f"Starting tunneling {self.__middle_host}:{self.__middle_port} => "
                                  f"{self.__host}:{self.__port}")
        with SSHTunnelForwarder(
                ssh_address_or_host=(self.__middle_host, self.__middle_port),
                ssh_username=self.__middle_username, ssh_password=self.__middle_password,
                remote_bind_address=(self.__host, self.__port),
                set_keepalive=100, threaded=False) as server:
            self.__ssh_host = server.local_bind_host
            self.__ssh_port = server.local_bind_port
            self.__log(logging.INFO, f"Started tunneling middle success "
                                     f"{self.__ssh_host}:{self.__ssh_port} => "
                                     f"{self.__host}:{self.__port}")
            self._set_connect(handle=handle, retry_count=retry_count, retry_delay=retry_delay, **kwargs)
            self.__log(logging.DEBUG, f"Closing tunneling "
                                      f"{self.__ssh_host}:{self.__ssh_port} => "
                                      f"{self.__host}:{self.__port}")
        return self.return_value



"""
    User
"""


def check_user_id(user_id: str):
    """
    For check user id string
    :param user_id:
    """
    if not user_id.startswith("@"):
        raise ValueError("UserIDs start with @")

    if ":" not in user_id:
        raise ValueError("UserIDs must have a domain component, separated by a :")


class User(object):
    """ The User class can be used to call user specific functions.
    """

    def __init__(self, api, user_id: str, display_name: str = None):
        check_user_id(user_id)

        self.user_id = user_id
        self.display_name = display_name
        self.api = api

    def get_display_name(self, room=None):
        """Get this user's display name.

        Args:
            room (Room): Optional. When specified, return the display name of the user
                in this room.

        Returns:
            The display name. Defaults to the user ID if not set.
        """
        if room:
            try:
                return room.members_display_names[self.user_id]
            except KeyError:
                return self.user_id
        if not self.display_name:
            self.display_name = self.api.get_display_name(self.user_id)
        return self.display_name or self.user_id

    def set_display_name(self, display_name):
        """ Set this users display name.

        Args:
            display_name (str): Display Name
        """
        self.display_name = display_name
        return self.api.set_display_name(self.user_id, display_name)



"""
    Error list
"""
# -*- coding: utf-8 -*-
# Copyright 2015 OpenMarket Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class MatrixError(Exception):
    """A generic Matrix error. Specific errors will subclass this."""
    pass


class MatrixUnexpectedResponse(MatrixError):
    """The home server gave an unexpected response. """

    def __init__(self, content=""):
        super(MatrixUnexpectedResponse, self).__init__(content)
        self.content = content


class MatrixRequestError(MatrixError):
    """ The home server returned an error response. """

    def __init__(self, code=0, content=""):
        super(MatrixRequestError, self).__init__("%d: %s" % (code, content))
        self.code = code
        self.content = content


class MatrixHttpLibError(MatrixError):
    """The library used for http requests raised an exception."""

    def __init__(self, original_exception, method, endpoint):
        super(MatrixHttpLibError, self).__init__(
            "Something went wrong in {} requesting {}: {}".format(method,
                                                                  endpoint,
                                                                  original_exception)
        )
        self.original_exception = original_exception



"""
    MatrixHttpApi Handler
"""
# -*- coding: utf-8 -*-
import json
import logging
from time import time, sleep

from requests import Session, RequestException, codes
from .errors import MatrixError, MatrixRequestError, MatrixHttpLibError

try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote

MATRIX_V2_API_PATH = "/_matrix/client/r0"

logger = logging.getLogger(__name__)


class MatrixHttpApi(object):
    """
        All API Functions
    """

    def __init__(
            self, base_url: str, token: str = None,
            default_429_wait_ms: int = 5000,
            connection_timeout: int = 60,
            validate_cert: bool = True
    ):
        self.base_url = base_url
        self.token = token
        self.txn_id = 0
        self.validate_cert = validate_cert
        self.session = Session()
        self.default_429_wait_ms = default_429_wait_ms
        self.connection_timeout = connection_timeout

    def sync(self, since=None, timeout_ms=30000, request_filter=None,
             full_state: bool = False, set_presence: str = None):
        """ Perform a sync request.

        Args:
            since (str): Optional. A token which specifies where to continue a sync from.
            timeout_ms (int): Optional. The time in milliseconds to wait.
            request_filter (int|str): Either a Filter ID or a JSON string.
            full_state (bool): Return the full state for every room the user has joined
                Defaults to false.
            set_presence (str): Should the client be marked as "online" or" offline"
        """

        request = {
            "timeout": int(timeout_ms)
        }

        if since:
            request["since"] = since

        if request_filter:
            request["filter"] = request_filter

        if full_state:
            request["full_state"] = json.dumps(full_state)

        if set_presence:
            request["set_presence"] = set_presence

        return self._send("GET", "/sync", query_params=request)

    def login(self, login_type, **kwargs):
        """Perform /login.

        Args:
            login_type (str): The value for the 'type' key.
            **kwargs: Additional key/values to add to the JSON submitted.
        """
        content = {
            "type": login_type
        }

        for key, value in kwargs.items():
            content[key] = value

        return self._send("POST", "/login", content)

    def logout(self):
        """Perform /logout.
        """
        return self._send("POST", "/logout")

    def logout_all(self):
        """Perform /logout/all."""
        return self._send("POST", "/logout/all")

    def create_room(
            self,
            alias=None,
            name=None,
            is_public=False,
            invitees=None,
            federate=True
    ):
        """Perform /createRoom.

        Args:
            alias (str): Optional. The room alias name to set for this room.
            name (str): Optional. Name for new room.
            is_public (bool): Optional. The public/private visibility.
            invitees (list<str>): Optional. The list of user IDs to invite.
            federate (bool): Optional. Сan a room be federated.
                Default to True.
        """
        content = {
            "visibility": "public" if is_public else "private"
        }
        if alias:
            content["room_alias_name"] = alias
        if invitees:
            content["invite"] = invitees
        if name:
            content["name"] = name
        if federate is not None:
            content["creation_content"] = {'m.federate': federate}
        return self._send("POST", "/createRoom", content)

    def join_room(self, room_id_or_alias):
        """Performs /join/$room_id

        Args:
            room_id_or_alias (str): The room ID or room alias to join.
        """
        if not room_id_or_alias:
            raise MatrixError("No alias or room ID to join.")

        path = "/join/%s" % quote(room_id_or_alias)

        return self._send("POST", path)

    def send_state_event(self, room_id, event_type, content, state_key="",
                         timestamp=None):
        """Perform PUT /rooms/$room_id/state/$event_type

        Args:
            room_id(str): The room ID to send the state event in.
            event_type(str): The state event type to send.
            content(dict): The JSON content to send.
            state_key(str): Optional. The state key for the event.
            timestamp (int): Set origin_server_ts (For application services only)
        """
        path = "/rooms/%s/state/%s" % (
            quote(room_id), quote(event_type),
        )
        if state_key:
            path += "/%s" % (quote(state_key))
        params = {}
        if timestamp:
            params["ts"] = timestamp
        return self._send("PUT", path, content, query_params=params)

    def get_state_event(self, room_id, event_type):
        """Perform GET /rooms/$room_id/state/$event_type

        Args:
            room_id(str): The room ID.
            event_type (str): The type of the event.

        Raises:
            MatrixRequestError(code=404) if the state event is not found.
        """
        return self._send("GET", "/rooms/{}/state/{}".format(quote(room_id), event_type))

    def send_message_event(self, room_id, event_type, content, txn_id=None,
                           timestamp=None):
        """Perform PUT /rooms/$room_id/send/$event_type

        Args:
            room_id (str): The room ID to send the message event in.
            event_type (str): The event type to send.
            content (dict): The JSON content to send.
            txn_id (int): Optional. The transaction ID to use.
            timestamp (int): Set origin_server_ts (For application services only)
        """
        if not txn_id:
            txn_id = self._make_txn_id()

        path = "/rooms/%s/send/%s/%s" % (
            quote(room_id), quote(event_type), quote(str(txn_id)),
        )
        params = {}
        if timestamp:
            params["ts"] = timestamp
        return self._send("PUT", path, content, query_params=params)

    def redact_event(self, room_id, event_id, reason=None, txn_id=None, timestamp=None):
        """Perform PUT /rooms/$room_id/redact/$event_id/$txn_id/

        Args:
            room_id(str): The room ID to redact the message event in.
            event_id(str): The event id to redact.
            reason (str): Optional. The reason the message was redacted.
            txn_id(int): Optional. The transaction ID to use.
            timestamp(int): Optional. Set origin_server_ts (For application services only)
        """
        if not txn_id:
            txn_id = self._make_txn_id()

        path = '/rooms/%s/redact/%s/%s' % (
            room_id, event_id, txn_id
        )
        content = {}
        if reason:
            content['reason'] = reason
        params = {}
        if timestamp:
            params["ts"] = timestamp
        return self._send("PUT", path, content, query_params=params)

    # content_type can be a image,audio or video
    # extra information should be supplied, see
    # https://matrix.org/docs/spec/r0.0.1/client_server.html
    def send_content(self, room_id, item_url, item_name, msg_type,
                     extra_information=None, timestamp=None):
        """
            Send Content
        """
        if extra_information is None:
            extra_information = {}

        content_pack = {
            "url": item_url,
            "msgtype": msg_type,
            "body": item_name,
            "info": extra_information
        }
        return self.send_message_event(room_id, "m.room.message", content_pack,
                                       timestamp=timestamp)

    # http://matrix.org/docs/spec/client_server/r0.2.0.html#m-location
    def send_location(self, room_id, geo_uri, name, thumb_url=None, thumb_info=None,
                      timestamp=None):
        """Send m.location message event

        Args:
            room_id (str): The room ID to send the event in.
            geo_uri (str): The geo uri representing the location.
            name (str): Description for the location.
            thumb_url (str): URL to the thumbnail of the location.
            thumb_info (dict): Metadata about the thumbnail, type ImageInfo.
            timestamp (int): Set origin_server_ts (For application services only)
        """
        content_pack = {
            "geo_uri": geo_uri,
            "msgtype": "m.location",
            "body": name,
        }
        if thumb_url:
            content_pack["thumbnail_url"] = thumb_url
        if thumb_info:
            content_pack["thumbnail_info"] = thumb_info

        return self.send_message_event(room_id, "m.room.message", content_pack,
                                       timestamp=timestamp)

    def send_message(self, room_id, text_content, message_type="m.text", timestamp=None):
        """Perform PUT /rooms/$room_id/send/m.room.message

        Args:
            message_type (str): ...
            room_id (str): The room ID to send the event in.
            text_content (str): The m.text body to send.
            timestamp (int): Set origin_server_ts (For application services only)
        """
        return self.send_message_event(
            room_id, "m.room.message",
            self.get_text_body(text_content, message_type),
            timestamp=timestamp
        )

    def send_emote(self, room_id, text_content, timestamp=None):
        """Perform PUT /rooms/$room_id/send/m.room.message with m.emote msgtype

        Args:
            room_id (str): The room ID to send the event in.
            text_content (str): The m.emote body to send.
            timestamp (int): Set origin_server_ts (For application services only)
        """
        return self.send_message_event(
            room_id, "m.room.message",
            self.get_emote_body(text_content),
            timestamp=timestamp
        )

    def send_notice(self, room_id, text_content, timestamp=None):
        """Perform PUT /rooms/$room_id/send/m.room.message with m.notice msgtype

        Args:
            room_id (str): The room ID to send the event in.
            text_content (str): The m.notice body to send.
            timestamp (int): Set origin_server_ts (For application services only)
        """
        body = {
            "msgtype": "m.notice",
            "body": text_content
        }
        return self.send_message_event(room_id, "m.room.message", body,
                                       timestamp=timestamp)

    def get_room_messages(self, room_id, token, direction, limit=10, to=None):
        """Perform GET /rooms/{roomId}/messages.

        Args:
            room_id (str): The room's id.
            token (str): The token to start returning events from.
            direction (str):  The direction to return events from. One of: ["b", "f"].
            limit (int): The maximum number of events to return.
            to (str): The token to stop returning events at.
        """
        query = {
            "roomId": room_id,
            "from": token,
            "dir": direction,
            "limit": limit,
        }

        if to:
            query["to"] = to

        return self._send("GET", "/rooms/{}/messages".format(quote(room_id)),
                          query_params=query, api_path="/_matrix/client/r0")

    def get_room_name(self, room_id):
        """Perform GET /rooms/$room_id/state/m.room.name
        Args:
            room_id(str): The room ID
        """
        return self.get_state_event(room_id, "m.room.name")

    def set_room_name(self, room_id, name, timestamp=None):
        """Perform PUT /rooms/$room_id/state/m.room.name
        Args:
            room_id (str): The room ID
            name (str): The new room name
            timestamp (int): Set origin_server_ts (For application services only)
        """
        body = {
            "name": name
        }
        return self.send_state_event(room_id, "m.room.name", body, timestamp=timestamp)

    def get_room_topic(self, room_id):
        """Perform GET /rooms/$room_id/state/m.room.topic
        Args:
            room_id (str): The room ID
        """
        return self.get_state_event(room_id, "m.room.topic")

    def set_room_topic(self, room_id, topic, timestamp=None):
        """Perform PUT /rooms/$room_id/state/m.room.topic
        Args:
            room_id (str): The room ID
            topic (str): The new room topic
            timestamp (int): Set origin_server_ts (For application services only)
        """
        body = {
            "topic": topic
        }
        return self.send_state_event(room_id, "m.room.topic", body, timestamp=timestamp)

    def get_power_levels(self, room_id):
        """Perform GET /rooms/$room_id/state/m.room.power_levels

        Args:
            room_id(str): The room ID
        """
        return self.get_state_event(room_id, "m.room.power_levels")

    def set_power_levels(self, room_id, content):
        """Perform PUT /rooms/$room_id/state/m.room.power_levels

        Note that any power levels which are not explicitly specified
        in the content arg are reset to default values.

        Args:
            room_id (str): The room ID
            content (dict): The JSON content to send. See example content below.

        Example::

            api = MatrixHttpApi("http://example.com", token="foobar")
            api.set_power_levels("!example:example.com",
                {
                    "ban": 50, # defaults to 50 if unspecified
                    "events": {
                        "m.room.name": 100, # must have PL 100 to change room name
                        "m.room.power_levels": 100 # must have PL 100 to change PLs
                    },
                    "events_default": 0, # defaults to 0
                    "invite": 50, # defaults to 50
                    "kick": 50, # defaults to 50
                    "redact": 50, # defaults to 50
                    "state_default": 50, # defaults to 50 if m.room.power_levels exists
                    "users": {
                        "@someone:example.com": 100 # defaults to 0
                    },
                    "users_default": 0 # defaults to 0
                }
            )
        """
        # Synapse returns M_UNKNOWN if body['events'] is omitted,
        #  as of 2016-10-31
        if "events" not in content:
            content["events"] = {}

        return self.send_state_event(room_id, "m.room.power_levels", content)

    def leave_room(self, room_id):
        """Perform POST /rooms/$room_id/leave

        Args:
            room_id (str): The room ID
        """
        return self._send("POST", "/rooms/" + room_id + "/leave", {})

    def forget_room(self, room_id):
        """Perform POST /rooms/$room_id/forget

        Args:
            room_id(str): The room ID
        """
        return self._send("POST", "/rooms/" + room_id + "/forget", content={})

    def invite_user(self, room_id, user_id):
        """Perform POST /rooms/$room_id/invite

        Args:
            room_id (str): The room ID
            user_id (str): The user ID of the invitee
        """
        body = {
            "user_id": user_id
        }
        return self._send("POST", "/rooms/" + room_id + "/invite", body)

    def kick_user(self, room_id, user_id, reason=""):
        """Calls set_membership with membership="leave" for the user_id provided
        """
        self.set_membership(room_id, user_id, "leave", reason)

    def get_membership(self, room_id, user_id):
        """Perform GET /rooms/$room_id/state/m.room.member/$user_id

        Args:
            room_id (str): The room ID
            user_id (str): The user ID
        """
        return self._send(
            "GET",
            "/rooms/%s/state/m.room.member/%s" % (room_id, user_id)
        )

    def set_membership(self, room_id, user_id, membership, reason="", profile=None,
                       timestamp=None):
        """
        Perform PUT /rooms/$room_id/state/m.room.member/$user_id

        """
        if profile is None:
            profile = {}
        body = {
            "membership": membership,
            "reason": reason
        }
        if 'displayname' in profile:
            body["displayname"] = profile["displayname"]
        if 'avatar_url' in profile:
            body["avatar_url"] = profile["avatar_url"]

        return self.send_state_event(room_id, "m.room.member", body, state_key=user_id,
                                     timestamp=timestamp)

    def ban_user(self, room_id, user_id, reason=""):
        """Perform POST /rooms/$room_id/ban

        Args:
            room_id (str): The room ID
            user_id (str): The user ID of the ban(sic)
            reason (str): The reason for this ban
        """
        body = {
            "user_id": user_id,
            "reason": reason
        }
        return self._send("POST", "/rooms/" + room_id + "/ban", body)

    def unban_user(self, room_id, user_id):
        """Perform POST /rooms/$room_id/unban

        Args:
            room_id (str): The room ID
            user_id (str): The user ID of the ban(sic)
        """
        body = {
            "user_id": user_id
        }
        return self._send("POST", "/rooms/" + room_id + "/unban", body)

    def get_user_tags(self, user_id, room_id):
        """
            get user tags
        """
        return self._send(
            "GET",
            "/user/%s/rooms/%s/tags" % (user_id, room_id),
        )

    def remove_user_tag(self, user_id, room_id, tag):
        """
            remove user tag
        """
        return self._send(
            "DELETE",
            "/user/%s/rooms/%s/tags/%s" % (user_id, room_id, tag),
        )

    def add_user_tag(self, user_id, room_id, tag, order=None, body=None):
        """
            add user tag
        """
        if body:
            pass
        elif order:
            body = {"order": order}
        else:
            body = {}
        return self._send(
            "PUT",
            "/user/%s/rooms/%s/tags/%s" % (user_id, room_id, tag),
            body,
        )

    def set_account_data(self, user_id, message_type, account_data):
        """
            set account data
        """
        return self._send(
            "PUT",
            "/user/%s/account_data/%s" % (user_id, message_type),
            account_data,
        )

    def set_room_account_data(self, user_id, room_id, message_type, account_data):
        """
            set room account data
        """
        return self._send(
            "PUT",
            "/user/%s/rooms/%s/account_data/%s" % (user_id, room_id, message_type),
            account_data
        )

    def get_room_state(self, room_id):
        """Perform GET /rooms/$room_id/state

        Args:
            room_id (str): The room ID
        """
        return self._send("GET", "/rooms/" + room_id + "/state")

    @staticmethod
    def get_text_body(text, message_type="m.text"):
        """
            Get text body
        """
        return {
            "msgtype": message_type,
            "body": text
        }

    @staticmethod
    def get_emote_body(text):
        """
            Get emote body
        """
        return {
            "msgtype": "m.emote",
            "body": text
        }

    def get_filter(self, user_id, filter_id):
        """
            Get filter
        """
        return self._send("GET", "/user/{userId}/filter/{filterId}"
                          .format(userId=user_id, filterId=filter_id))

    def create_filter(self, user_id, filter_params):
        """
            Create filter
        """
        return self._send("POST",
                          "/user/{userId}/filter".format(userId=user_id),
                          filter_params)

    def _send(self, method, path, content=None, query_params=None, headers=None,
              api_path=MATRIX_V2_API_PATH, return_json=True):
        if query_params is None:
            query_params = {}
        if headers is None:
            headers = {}

        if "User-Agent" not in headers:
            headers["User-Agent"] = "seraj_server"

        method = method.upper()
        if method not in ["GET", "PUT", "DELETE", "POST"]:
            raise MatrixError("Unsupported HTTP method: %s" % method)

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"

        query_params["access_token"] = self.token

        endpoint = self.base_url + api_path + path

        if headers["Content-Type"] == "application/json" and content is not None:
            content = json.dumps(content)
        while True:
            try:
                response = self.session.request(
                    method, endpoint,
                    params=query_params,
                    data=content,
                    headers=headers,
                    verify=self.validate_cert,
                    timeout=self.connection_timeout,
                )
            except RequestException as e:
                raise MatrixHttpLibError(e, method, endpoint)
            try:
                json_response = response.json()
            except Exception as e:
                logger.error(e)
                json_response = {}
            if response.status_code == codes.too_many_requests:
                wait_time = json_response.get('retry_after_ms', self.default_429_wait_ms / 1000)
                if json_response.get('error') is not None:
                    logger.error(json_response.get('error'))
                sleep(wait_time)
            else:
                break

        if response.status_code < codes.ok or response.status_code >= codes.multiple_choices:
            raise MatrixRequestError(
                code=response.status_code, content=response.text
            )
        if return_json:
            return json_response
        else:
            return response

    def media_upload(self, content, content_type, filename=None):
        """
            set room account data
        """
        query_params = {}
        if filename is not None:
            query_params['filename'] = filename

        return self._send(
            "POST", "",
            content=content,
            headers={"Content-Type": content_type},
            api_path="/_matrix/media/r0/upload",
            query_params=query_params
        )

    def get_display_name(self, user_id):
        """
            get display name
        """
        content = self._send("GET", "/profile/%s/displayname" % user_id)
        return content.get('displayname', None)

    def set_display_name(self, user_id, display_name):
        """
            set display name
        """
        content = {"displayname": display_name}
        return self._send("PUT", "/profile/%s/displayname" % user_id, content)

    def get_avatar_url(self, user_id):
        """
            get avatar url
        """
        content = self._send("GET", "/profile/%s/avatar_url" % user_id)
        return content.get('avatar_url', None)

    def set_avatar_url(self, user_id, avatar_url):
        """
            set avatar url
        """
        content = {"avatar_url": avatar_url}
        return self._send("PUT", "/profile/%s/avatar_url" % user_id, content)

    def get_download_url(self, matrix_url):
        """
            get download url
        """
        if matrix_url.startswith('mxc://'):
            return self.base_url + "/_matrix/media/r0/download/" + matrix_url[6:]
        else:
            raise ValueError("MXC URL did not begin with 'mxc://'")

    def media_download(self, matrix_url, allow_remote=True):
        """Download raw media from provided mxc URL.

        Args:
            matrix_url (str): mxc media URL.
            allow_remote (bool): indicates to the server that it should not
                attempt to fetch the media if it is deemed remote. Defaults
                to true if not provided.
        """
        query_params = {}
        if not allow_remote:
            query_params["allow_remote"] = False
        if matrix_url.startswith('mxc://'):
            return self._send(
                "GET", matrix_url[6:],
                api_path="/_matrix/media/r0/download/",
                query_params=query_params,
                return_json=False
            )
        else:
            raise ValueError(
                "MXC URL '%s' did not begin with 'mxc://'" % matrix_url
            )

    def get_thumbnail(self, matrix_url, width, height, method='scale', allow_remote=True):
        """Download raw media thumbnail from provided mxc URL.

        Args:
            matrix_url (str): mxc media URL
            width (int): desired thumbnail width
            height (int): desired thumbnail height
            method (str): thumb creation method. Must be
                in ['scale', 'crop']. Default 'scale'.
            allow_remote (bool): indicates to the server that it should not
                attempt to fetch the media if it is deemed remote. Defaults
                to true if not provided.
        """
        if method not in ['scale', 'crop']:
            raise ValueError(
                "Unsupported thumb method '%s'" % method
            )
        query_params = {
            "width": width,
            "height": height,
            "method": method
        }
        if not allow_remote:
            query_params["allow_remote"] = False
        if matrix_url.startswith('mxc://'):
            return self._send(
                "GET", matrix_url[6:],
                query_params=query_params,
                api_path="/_matrix/media/r0/thumbnail/",
                return_json=False
            )
        else:
            raise ValueError(
                "MXC URL '%s' did not begin with 'mxc://'" % matrix_url
            )

    def get_url_preview(self, url, ts=None):
        """Get preview for URL.

        Args:
            url (str): URL to get a preview
            ts (double): The preferred point in time to return
                 a preview for. The server may return a newer
                 version if it does not have the requested
                 version available.
        """
        params = {'url': url}
        if ts:
            params['ts'] = ts
        return self._send(
            "GET", "",
            query_params=params,
            api_path="/_matrix/media/r0/preview_url"
        )

    def get_room_id(self, room_alias):
        """Get room id from its alias.

        Args:
            room_alias (str): The room alias name.

        Returns:
            Wanted room's id.
        """
        content = self._send("GET", "/directory/room/{}".format(quote(room_alias)))
        return content.get("room_id", None)

    def set_room_alias(self, room_id, room_alias):
        """Set alias to room id

        Args:
            room_id (str): The room id.
            room_alias (str): The room wanted alias name.
        """
        data = {
            "room_id": room_id
        }

        return self._send("PUT", "/directory/room/{}".format(quote(room_alias)),
                          content=data)

    def remove_room_alias(self, room_alias):
        """Remove mapping of an alias

        Args:
            room_alias(str): The alias to be removed.

        Raises:
            MatrixRequestError
        """
        return self._send("DELETE", "/directory/room/{}".format(quote(room_alias)))

    def get_room_members(self, room_id):
        """Get the list of members for this room.

        Args:
            room_id (str): The room to get the member events for.
        """
        return self._send("GET", "/rooms/{}/members".format(quote(room_id)))

    def set_join_rule(self, room_id, join_rule):
        """Set the rule for users wishing to join the room.

        Args:
            room_id(str): The room to set the rules for.
            join_rule(str): The chosen rule. One of: ["public", "knock",
                "invite", "private"]
        """
        content = {
            "join_rule": join_rule
        }
        return self.send_state_event(room_id, "m.room.join_rules", content)

    def set_guest_access(self, room_id, guest_access):
        """Set the guest access policy of the room.

        Args:
            room_id(str): The room to set the rules for.
            guest_access(str): Whether guests can join. One of: ["can_join",
                "forbidden"]
        """
        content = {
            "guest_access": guest_access
        }
        return self.send_state_event(room_id, "m.room.guest_access", content)

    def get_devices(self):
        """Gets information about all devices for the current user."""
        return self._send("GET", "/devices")

    def get_device(self, device_id):
        """Gets information on a single device, by device id."""
        return self._send("GET", "/devices/%s" % device_id)

    def update_device_info(self, device_id, display_name):
        """Update the display name of a device.

        Args:
            device_id (str): The device ID of the device to update.
            display_name (str): New display name for the device.
        """
        content = {
            "display_name": display_name
        }
        return self._send("PUT", "/devices/%s" % device_id, content=content)

    def delete_device(self, auth_body, device_id):
        """Deletes the given device, and invalidates any access token associated with it.

        NOTE: This endpoint uses the User-Interactive Authentication API.

        Args:
            auth_body (dict): Authentication params.
            device_id (str): The device ID of the device to delete.
        """
        content = {
            "auth": auth_body
        }
        return self._send("DELETE", "/devices/%s" % device_id, content=content)

    def delete_devices(self, auth_body, devices):
        """Bulk deletion of devices.

        NOTE: This endpoint uses the User-Interactive Authentication API.

        Args:
            auth_body (dict): Authentication params.
            devices (list): List of device ID"s to delete.
        """
        content = {
            "auth": auth_body,
            "devices": devices
        }
        return self._send("POST", "/delete_devices", content=content)

    def upload_keys(self, device_keys=None, one_time_keys=None):
        """Publishes end-to-end encryption keys for the device.

        Said device must be the one used when logging in.

        Args:
            device_keys (dict): Optional. Identity keys for the device. The required
                keys are:

                | user_id (str): The ID of the user the device belongs to. Must match
                    the user ID used when logging in.
                | device_id (str): The ID of the device these keys belong to. Must match
                    the device ID used when logging in.
                | algorithms (list<str>): The encryption algorithms supported by this
                    device.
                | keys (dict): Public identity keys. Should be formatted as
                    <algorithm:device_id>: <key>.
                | signatures (dict): Signatures for the device key object. Should be
                    formatted as <user_id>: {<algorithm:device_id>: <key>}

            one_time_keys (dict): Optional. One-time public keys. Should be
                formatted as <algorithm:key_id>: <key>, the key format being
                determined by the algorithm.
        """
        content = {}
        if device_keys:
            content["device_keys"] = device_keys
        if one_time_keys:
            content["one_time_keys"] = one_time_keys
        return self._send("POST", "/keys/upload", content=content)

    def query_keys(self, user_devices, timeout=None, token=None):
        """Query HS for public keys by user and optionally device.

        Args:
            user_devices (dict): The devices whose keys to download. Should be
                formatted as <user_id>: [<device_ids>]. No device_ids indicates
                all devices for the corresponding user.
            timeout (int): Optional. The time (in milliseconds) to wait when
                downloading keys from remote servers.
            token (str): Optional. If the client is fetching keys as a result of
                a device update received in a sync request, this should be the
                'since' token of that sync request, or any later sync token.
        """
        content = {"device_keys": user_devices}
        if timeout:
            content["timeout"] = timeout
        if token:
            content["token"] = token
        return self._send("POST", "/keys/query", content=content)

    def claim_keys(self, key_request, timeout=None):
        """Claims one-time keys for use in pre-key messages.

        Args:
            key_request (dict): The keys to be claimed. Format should be
                <user_id>: { <device_id>: <algorithm> }.
            timeout (int): Optional. The time (in milliseconds) to wait when
                downloading keys from remote servers.
        """
        content = {"one_time_keys": key_request}
        if timeout:
            content["timeout"] = timeout
        return self._send("POST", "/keys/claim", content=content)

    def key_changes(self, from_token, to_token):
        """Gets a list of users who have updated their device identity keys.

        Args:
            from_token (str): The desired start point of the list. Should be the
                next_batch field from a response to an earlier call to /sync.
            to_token (str): The desired end point of the list. Should be the next_batch
                field from a recent call to /sync - typically the most recent such call.
        """
        params = {"from": from_token, "to": to_token}
        return self._send("GET", "/keys/changes", query_params=params)

    def send_to_device(self, event_type, messages, txn_id=None):
        """Sends send-to-device events to a set of client devices.

        Args:
            event_type (str): The type of event to send.
            messages (dict): The messages to send. Format should be
                <user_id>: {<device_id>: <event_content>}.
                The device ID may also be '*', meaning all known devices for the user.
            txn_id (str): Optional. The transaction ID for this event, will be generated
                automatically otherwise.
        """
        txn_id = txn_id if txn_id else self._make_txn_id()
        return self._send(
            "PUT",
            "/sendToDevice/{}/{}".format(event_type, txn_id),
            content={"messages": messages}
        )

    def _make_txn_id(self):
        txn_id = str(self.txn_id) + str(int(time() * 1000))
        self.txn_id += 1
        return txn_id

    def whoami(self):
        """
            Determine user_id for authenticated user.
        """
        if not self.token:
            raise MatrixError("Authentication required.")
        return self._send(
            "GET",
            "/account/whoami"
        )



"""
    Room
"""
import logging
import re
from uuid import uuid4

from .errors import MatrixRequestError
from .user import User

logger = logging.getLogger(__name__)


def check_room_id(room_id):
    """
        Check room id
    """
    if not room_id.startswith("!"):
        raise ValueError("RoomIDs start with !")

    if ":" not in room_id:
        raise ValueError("RoomIDs must have a domain component, separated by a :")


class Room(object):
    """
        Call room-specific functions after joining a room from the client.
    """

    def __init__(self, client, room_id):
        check_room_id(room_id)

        self.room_id = room_id
        self.client = client
        self.listeners = []
        self.state_listeners = []
        self.ephemeral_listeners = []
        self.events = []
        self.event_history_limit = 20
        self.name = None
        self.canonical_alias = None
        self.aliases = []
        self.topic = None
        self.invite_only = None
        self.guest_access = None
        self._prev_batch = None
        self._members = {}
        self.members_display_names = {
            # user_id: displayname
        }
        self.encrypted = False

    def set_user_profile(self,
                         displayname=None,
                         avatar_url=None,
                         reason="Changing room profile information"):
        """Set user profile within a room.

        This sets displayname and avatar_url for the logged in user only in a
        specific room. It does not change the user's global user profile.
        """
        member = self.client.api.get_membership(self.room_id, self.client.user_id)
        if member["membership"] != "join":
            raise Exception("Can't set profile if you have not joined the room.")
        if displayname is None:
            displayname = member["displayname"]
        if avatar_url is None:
            avatar_url = member["avatar_url"]
        self.client.api.set_membership(
            self.room_id,
            self.client.user_id,
            'join',
            reason, {
                "displayname": displayname,
                "avatar_url": avatar_url
            }
        )

    @property
    def display_name(self):
        """Calculates the display name for a room."""
        if self.name:
            return self.name
        elif self.canonical_alias:
            return self.canonical_alias

        # Member display names without me
        members = [u.get_display_name(self) for u in self.get_joined_members() if
                   self.client.user_id != u.user_id]
        members.sort()

        if len(members) == 1:
            return members[0]
        elif len(members) == 2:
            return "{0} and {1}".format(members[0], members[1])
        elif len(members) > 2:
            return "{0} and {1} others".format(members[0], len(members) - 1)
        else:  # len(members) <= 0 or not an integer
            # TODO i18n
            return "Empty room"

    def send_text(self, text):
        """Send a plain text message to the room."""
        return self.client.api.send_message(self.room_id, text)

    # noinspection PyMissingOrEmptyDocstring
    @staticmethod
    def get_html_content(html, body=None, message_type="m.text"):
        return {
            "body": body if body else re.sub('<[^<]+?>', '', html),
            "msgtype": message_type,
            "format": "org.matrix.custom.html",
            "formatted_body": html
        }

    def send_html(self, html, body: str = None, message_type: str = "m.text"):
        """Send an html formatted message.

        Args:
            message_type (str): ...
            html (str): The html formatted message to be sent.
            body (str): The un_formatted body of the message to be sent.
        """
        return self.client.api.send_message_event(
            self.room_id, "m.room.message", self.get_html_content(html, body, message_type))

    # noinspection PyMissingOrEmptyDocstring
    def set_account_data(self, message_type, account_data):
        return self.client.api.set_room_account_data(
            self.client.user_id, self.room_id, message_type, account_data)

    # noinspection PyMissingOrEmptyDocstring
    def get_tags(self):
        return self.client.api.get_user_tags(self.client.user_id, self.room_id)

    # noinspection PyMissingOrEmptyDocstring
    def remove_tag(self, tag):
        return self.client.api.remove_user_tag(
            self.client.user_id, self.room_id, tag
        )

    # noinspection PyMissingOrEmptyDocstring
    def add_tag(self, tag, order=None, content=None):
        return self.client.api.add_user_tag(
            self.client.user_id, self.room_id,
            tag, order, content
        )

    def send_emote(self, text):
        """Send an emote (/me style) message to the room."""
        return self.client.api.send_emote(self.room_id, text)

    def send_file(self, url, name, **file_info):
        """Send a pre-uploaded file to the room.

        See http://matrix.org/docs/spec/r0.2.0/client_server.html#m-file for
        file_info.

        Args:
            url (str): The mxc url of the file.
            name (str): The filename of the image.
            file_info (): Extra information about the file
        """

        return self.client.api.send_content(
            self.room_id, url, name, "m.file",
            extra_information=file_info
        )

    def send_notice(self, text):
        """Send a notice (from bot) message to the room."""
        return self.client.api.send_notice(self.room_id, text)

    # See http://matrix.org/docs/spec/r0.0.1/client_server.html#m-image for the
    # image_info args.
    def send_image(self, url, name, **image_info):
        """Send a pre-uploaded image to the room.

        See http://matrix.org/docs/spec/r0.0.1/client_server.html#m-image
        for image_info

        Args:
            url (str): The mxc url of the image.
            name (str): The filename of the image.
            image_info (): Extra information about the image.
        """
        return self.client.api.send_content(
            self.room_id, url, name, "m.image",
            extra_information=image_info
        )

    def send_location(self, geo_uri, name, thumb_url=None, **thumb_info):
        """Send a location to the room.

        See http://matrix.org/docs/spec/client_server/r0.2.0.html#m-location
        for thumb_info

        Args:
            geo_uri (str): The geo uri representing the location.
            name (str): Description for the location.
            thumb_url (str): URL to the thumbnail of the location.
            thumb_info (): Metadata about the thumbnail, type ImageInfo.
        """
        return self.client.api.send_location(self.room_id, geo_uri, name,
                                             thumb_url, thumb_info)

    def send_video(self, url, name, **video_info):
        """Send a pre-uploaded video to the room.

        See http://matrix.org/docs/spec/client_server/r0.2.0.html#m-video
        for video_info

        Args:
            url (str): The mxc url of the video.
            name (str): The filename of the video.
            video_info (): Extra information about the video.
        """
        return self.client.api.send_content(self.room_id, url, name, "m.video",
                                            extra_information=video_info)

    def send_audio(self, url, name, **audio_info):
        """Send a pre-uploaded audio to the room.

        See http://matrix.org/docs/spec/client_server/r0.2.0.html#m-audio
        for audio_info

        Args:
            url (str): The mxc url of the audio.
            name (str): The filename of the audio.
            audio_info (): Extra information about the audio.
        """
        return self.client.api.send_content(self.room_id, url, name, "m.audio",
                                            extra_information=audio_info)

    def redact_message(self, event_id, reason=None):
        """Redacts the message with specified event_id for the given reason.

        See https://matrix.org/docs/spec/r0.0.1/client_server.html#id112
        """
        return self.client.api.redact_event(self.room_id, event_id, reason)

    def add_listener(self, callback, event_type=None):
        """Add a callback handler for events going to this room.

        Args:
            callback (func(room, event)): Callback called when an event arrives.
            event_type (str): The event_type to filter for.
        Returns:
            uuid.UUID: Unique id of the listener, can be used to identify the listener.
        """
        listener_id = uuid4()
        self.listeners.append(
            {
                'uid': listener_id,
                'callback': callback,
                'event_type': event_type
            }
        )
        return listener_id

    def remove_listener(self, uid):
        """Remove listener with given uid."""
        self.listeners[:] = (listener for listener in self.listeners
                             if listener['uid'] != uid)

    def add_ephemeral_listener(self, callback, event_type=None):
        """Add a callback handler for ephemeral events going to this room.

        Args:
            callback (func(room, event)): Callback called when an ephemeral event arrives.
            event_type (str): The event_type to filter for.
        Returns:
            uuid.UUID: Unique id of the listener, can be used to identify the listener.
        """
        listener_id = uuid4()
        self.ephemeral_listeners.append(
            {
                'uid': listener_id,
                'callback': callback,
                'event_type': event_type
            }
        )
        return listener_id

    def remove_ephemeral_listener(self, uid):
        """Remove ephemeral listener with given uid."""
        self.ephemeral_listeners[:] = (listener for listener in self.ephemeral_listeners
                                       if listener['uid'] != uid)

    def add_state_listener(self, callback, event_type=None):
        """Add a callback handler for state events going to this room.

        Args:
            callback (func(room_chunk)): Callback called when an event arrives.
            event_type (str): The event_type to filter for.
        """
        self.state_listeners.append(
            {
                'callback': callback,
                'event_type': event_type
            }
        )

    def _put_event(self, event):
        self.events.append(event)
        if len(self.events) > self.event_history_limit:
            self.events.pop(0)
        if 'state_key' in event:
            self._process_state_event(event)

        # Dispatch for room-specific listeners
        for listener in self.listeners:
            if listener['event_type'] is None or listener['event_type'] == event['type']:
                listener['callback'](self, event)

    def _put_ephemeral_event(self, event):
        # Dispatch for room-specific listeners
        for listener in self.ephemeral_listeners:
            if listener['event_type'] is None or listener['event_type'] == event['type']:
                listener['callback'](self, event)

    def get_events(self):
        """Get the most recent events for this room."""
        return self.events

    def invite_user(self, user_id):
        """Invite a user to this room.

        Returns:
            boolean: Whether invitation was sent.
        """
        try:
            self.client.api.invite_user(self.room_id, user_id)
            return True
        except MatrixRequestError:
            return False

    def kick_user(self, user_id):
        """Kick a user from this room.


        Args:
            user_id (str): The matrix user id of a user.

        Returns:
            boolean: Whether user was kicked.
        """
        try:
            self.client.api.kick_user(self.room_id, user_id)
            return True
        except MatrixRequestError:
            return False

    def ban_user(self, user_id, reason):
        """Ban a user from this room

        Args:
            user_id (str): The matrix user id of a user.
            reason  (str): A reason for banning the user.

        Returns:
            boolean: The user was banned.
        """
        try:
            self.client.api.ban_user(self.room_id, user_id, reason)
            return True
        except MatrixRequestError:
            return False

    def unban_user(self, user_id):
        """Unban a user from this room

        Returns:
            boolean: The user was unbanned.
        """
        try:
            self.client.api.unban_user(self.room_id, user_id)
            return True
        except MatrixRequestError:
            return False

    def leave(self):
        """Leave the room.

        Returns:
            boolean: Leaving the room was successful.
        """
        try:
            self.client.api.leave_room(self.room_id)
            del self.client.rooms[self.room_id]
            return True
        except MatrixRequestError:
            return False

    def update_room_name(self):
        """Updates self.name and returns True if room name has changed."""
        try:
            response = self.client.api.get_room_name(self.room_id)
            if "name" in response and response["name"] != self.name:
                self.name = response["name"]
                return True
            else:
                return False
        except MatrixRequestError:
            return False

    def set_room_name(self, name):
        """Return True if room name successfully changed."""
        try:
            self.client.api.set_room_name(self.room_id, name)
            self.name = name
            return True
        except MatrixRequestError:
            return False

    def send_state_event(self, event_type, content, state_key=""):
        """Send a state event to the room.

        Args:
            event_type (str): The type of event that you are sending.
            content (): An object with the content of the message.
            state_key (str, optional): A unique key to identify the state.
        """
        return self.client.api.send_state_event(
            self.room_id,
            event_type,
            content,
            state_key
        )

    def update_room_topic(self):
        """Updates self.topic and returns True if room topic has changed."""
        try:
            response = self.client.api.get_room_topic(self.room_id)
            if "topic" in response and response["topic"] != self.topic:
                self.topic = response["topic"]
                return True
            else:
                return False
        except MatrixRequestError:
            return False

    def set_room_topic(self, topic):
        """Set room topic.

        Returns:
            boolean: True if the topic changed, False if not
        """
        try:
            self.client.api.set_room_topic(self.room_id, topic)
            self.topic = topic
            return True
        except MatrixRequestError:
            return False

    def update_aliases(self):
        """Get aliases information from room state.

        Returns:
            boolean: True if the aliases changed, False if not
        """
        try:
            response = self.client.api.get_room_state(self.room_id)
            for chunk in response:
                if "content" in chunk and "aliases" in chunk["content"]:
                    if chunk["content"]["aliases"] != self.aliases:
                        self.aliases = chunk["content"]["aliases"]
                        return True
                    else:
                        return False
        except MatrixRequestError:
            return False

    def add_room_alias(self, room_alias):
        """Add an alias to the room and return True if successful."""
        try:
            self.client.api.set_room_alias(self.room_id, room_alias)
            return True
        except MatrixRequestError:
            return False

    def get_joined_members(self):
        """Returns list of joined members (User objects)."""
        if self._members:
            return list(self._members.values())
        response = self.client.api.get_room_members(self.room_id)
        for event in response["chunk"]:
            if event["content"]["membership"] == "join":
                user_id = event["state_key"]
                self._add_member(user_id, event["content"].get("displayname"))
        return list(self._members.values())

    def _add_member(self, user_id, displayname=None):
        if displayname:
            self.members_display_names[user_id] = displayname
        if user_id in self._members:
            return
        if user_id in self.client.users:
            self._members[user_id] = self.client.users[user_id]
            return
        self._members[user_id] = User(self.client.api, user_id, displayname)
        self.client.users[user_id] = self._members[user_id]

    def back_fill_previous_messages(self, reverse=False, limit=10):
        """Back_fill handling of previous messages.

        Args:
            reverse (bool): When false messages will be back_filled in their original
                order (old to new), otherwise the order will be reversed (new to old).
            limit (int): Number of messages to go back.
        """
        res = self.client.api.get_room_messages(self.room_id, self.prev_batch,
                                                direction="b", limit=limit)
        events = res["chunk"]
        if not reverse:
            events = reversed(events)
        for event in events:
            self._put_event(event)

    def modify_user_power_levels(self, users=None, users_default=None):
        """Modify the power level for a subset of users

        Args:
            users(dict): Power levels to assign to specific users, in the form
                {"@name0:host0": 10, "@name1:host1": 100, "@name3:host3", None}
                A level of None causes the user to revert to the default level
                as specified by users_default.
            users_default(int): Default power level for users in the room

        Returns:
            True if successful, False if not
        """
        try:
            content = self.client.api.get_power_levels(self.room_id)
            if users_default:
                content["users_default"] = users_default

            if users:
                if "users" in content:
                    content["users"].update(users)
                else:
                    content["users"] = users

                # Remove any keys with value None
                for user, power_level in list(content["users"].items()):
                    if power_level is None:
                        del content["users"][user]
            self.client.api.set_power_levels(self.room_id, content)
            return True
        except MatrixRequestError:
            return False

    def modify_required_power_levels(self, events=None, **kwargs):
        """Modifies room power level requirements.

        Args:
            events(dict): Power levels required for sending specific event types,
                in the form {"m.room.whatever0": 60, "m.room.whatever2": None}.
                Overrides events_default and state_default for the specified
                events. A level of None causes the target event to revert to the
                default level as specified by events_default or state_default.
            **kwargs: Key/value pairs specifying the power levels required for
                    various actions:

                    - events_default(int): Default level for sending message events
                    - state_default(int): Default level for sending state events
                    - invite(int): Inviting a user
                    - redact(int): Redacting an event
                    - ban(int): Banning a user
                    - kick(int): Kicking a user

        Returns:
            True if successful, False if not
        """
        try:
            content = self.client.api.get_power_levels(self.room_id)
            content.update(kwargs)
            for key, value in list(content.items()):
                if value is None:
                    del content[key]

            if events:
                if "events" in content:
                    content["events"].update(events)
                else:
                    content["events"] = events

                # Remove any keys with value None
                for event, power_level in list(content["events"].items()):
                    if power_level is None:
                        del content["events"][event]

            self.client.api.set_power_levels(self.room_id, content)
            return True
        except MatrixRequestError:
            return False

    def set_invite_only(self, invite_only):
        """Set how the room can be joined.

        Args:
            invite_only(bool): If True, users will have to be invited to join
                the room. If False, anyone who knows the room link can join.

        Returns:
            True if successful, False if not
        """
        join_rule = "invite" if invite_only else "public"
        try:
            self.client.api.set_join_rule(self.room_id, join_rule)
            self.invite_only = invite_only
            return True
        except MatrixRequestError:
            return False

    def set_guest_access(self, allow_guests):
        """Set whether guests can join the room and return True if successful."""
        guest_access = "can_join" if allow_guests else "forbidden"
        try:
            self.client.api.set_guest_access(self.room_id, guest_access)
            self.guest_access = allow_guests
            return True
        except MatrixRequestError:
            return False

    def enable_encryption(self):
        """Enables encryption in the room.

        NOTE: Once enabled, encryption cannot be disabled.

        Returns:
        True if successful, False if not
        """
        try:
            self.send_state_event("m.room.encryption",
                                  {"algorithm": "m.megolm.v1.aes-sha2"})
            self.encrypted = True
            return True
        except MatrixRequestError:
            return False

    def _process_state_event(self, state_event):
        if "type" not in state_event:
            return  # Ignore event
        event_type = state_event["type"]
        event_content = state_event["content"]
        cache_level = 0

        # print(state_event)

        # Don't keep track of room state if caching turned off
        if cache_level >= 0:
            try:
                if event_type == "m.room.name":
                    self.name = event_content.get("name")
                elif event_type == "m.room.canonical_alias":
                    self.canonical_alias = event_content.get("alias")
                elif event_type == "m.room.topic":
                    self.topic = event_content.get("topic")
                elif event_type == "m.room.aliases":
                    self.aliases = event_content.get("aliases")
                elif event_type == "m.room.join_rules":
                    self.invite_only = event_content.get("join_rule") == "invite"
                elif event_type == "m.room.guest_access":
                    self.guest_access = event_content.get("guest_access") == "can_join"
                elif event_type == "m.room.encryption":
                    if event_content.get("algorithm") == "m.megolm.v1.aes-sha2":
                        self.encrypted = True
                elif event_type == "m.room.member":
                    # tracking room members can be large e.g. #matrix:matrix.org
                    if event_content["membership"] == "join":
                        user_id = state_event["state_key"]
                        self._add_member(user_id, event_content.get("displayname"))
                    elif event_content["membership"] in ("leave", "kick", "invite"):
                        self._members.pop(state_event["state_key"], None)
            except KeyError:
                logger.exception("Unable to parse state event %s, passing over.",
                                 state_event['event_id'])

        for listener in self.state_listeners:
            if (
                    listener['event_type'] is None or
                    listener['event_type'] == state_event['type']
            ):
                listener['callback'](state_event)

    # noinspection PyMissingOrEmptyDocstring
    @property
    def prev_batch(self):
        return self._prev_batch

    @prev_batch.setter
    def prev_batch(self, prev_batch):
        self._prev_batch = prev_batch



import json

from requests import Session


class BaseApi:
    """
        Use this class for handle all api
        Attributes:
            base_url (str) : base url
            token (str) : request token
            default_429_wait_time (int) : if response code is 429 wait this time
            connection_timeout (int) : wait this time for connect
            validate_cert (bool) : validate certificate for https connection
            user_agent (str) : default user agent for connection
            retry_count (int) : default retry count when error acquire
            retry_delay (int) : time to wait in error condition for next try
    """

    def __init__(self, base_url: str, token: str = None, default_429_wait_time: int = 5,
                 connection_timeout: int = 60, validate_cert: bool = True, user_agent: str = None,
                 retry_count: int = 1, retry_delay: int = 3):
        self.base_url = base_url
        self.token = token
        self.validate_cert = validate_cert
        self.session = Session()
        self.default_429_wait_time = default_429_wait_time
        self.connection_timeout = connection_timeout
        self.user_agent = user_agent
        self.retry_count = retry_count
        self.retry_delay = retry_delay

    def _send(self, method: str, path:str, api_path: str = "", content:dict=None, query_params:dict=None, headers=None, return_json=True):
        """
            Same function for send api request
        Args:
            method (str):
            path (str):
            api_path:
            content:
            query_params:
            headers:
            return_json:

        Returns:

        """
        if query_params is None:
            query_params = {}
        if headers is None:
            headers = {}

        if "User-Agent" not in headers and self.user_agent:
            headers["User-Agent"] = self.user_agent

        method = method.upper()
        if method not in ["GET", "PUT", "DELETE", "POST", "PATCH"]:
            raise Exception(f"Unsupported HTTP method: {method}")

        if "Content-Type" not in headers:
            headers["Content-Type"] = "application/json"

        query_params["access_token"] = self.token

        endpoint = self.base_url + api_path + path

        if headers["Content-Type"] == "application/json" and content is not None:
            content = json.dumps(content)
        while True:
            try:
                response = self.session.request(
                    method, endpoint,
                    params=query_params,
                    data=content,
                    headers=headers,
                    verify=self.validate_cert,
                    timeout=self.connection_timeout,
                )
            except RequestException as e:
                raise MatrixHttpLibError(e, method, endpoint)
            try:
                json_response = response.json()
            except Exception as e:
                logger.error(e)
                json_response = {}
            if response.status_code == codes.too_many_requests:
                wait_time = json_response.get('retry_after_ms', self.default_429_wait_ms / 1000)
                if json_response.get('error') is not None:
                    logger.error(json_response.get('error'))
                sleep(wait_time)
            else:
                break

        if response.status_code < codes.ok or response.status_code >= codes.multiple_choices:
            raise MatrixRequestError(
                code=response.status_code, content=response.text
            )
        if return_json:
            return json_response
        else:
            return response



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autoutils', 'autoutils.matrixcli']

package_data = \
{'': ['*']}

install_requires = \
['blessings>=1.0,<2.0',
 'func-timeout>=4.0,<5.0',
 'paramiko>=2.0,<3.0',
 'persiantools>=2.0,<3.0',
 'pytz>=2022.1,<2023.0',
 'redis>=4.0,<5.0',
 'requests>=2.0,<3.0',
 'sshtunnel>=0.1,<1.0',
 'urllib3>=1.0,<2.0']

setup_kwargs = {
    'name': 'autoutils',
    'version': '0.3.0',
    'description': 'Some Good Function',
    'long_description': '# autoutils',
    'author': 'Reza Zeiny',
    'author_email': 'rezazeiny1998@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/rezazeiny/autoutils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
    All Work with linux scripting
"""
__author__ = ('Reza Zeiny <rezazeiny1998@gmail.com>',)

import logging
import random
import string
import subprocess
from threading import Thread

logger = logging.getLogger(__name__)

PIPE = -1
STDOUT = -2
DEVNULL = -3


def id_generator(size: int = 6, chars: list = string.ascii_uppercase + string.ascii_lowercase + string.digits) -> str:
    """
        Generate random string
    Args:
        size (int): size of random string
        chars (list): define select set

    Returns:
        (str) : random string
    """
    return ''.join(random.choice(chars) for _ in range(size))


class ShellScript:
    """
        Shell Script
    """

    def __init__(self, command, *, shell=False, text=True, split_stdout=True, split_stderr=True, timeout=None,
                 background=False, quiet=False, extra_data=None):
        self.__used = False
        self.command = command
        self.is_running = False
        self.process_id = None
        self.exit_code = None
        self.stdout = None
        self.stderr = None
        self.error = None

        self.shell = shell
        self.split_stdout = split_stdout
        self.split_stderr = split_stderr
        self.text = text
        self.timeout = timeout
        self.background = background
        self.quiet = quiet
        self.extra_data = extra_data

    def _log(self, level, message, data=None):
        if data is None:
            data = {}
        if logger.isEnabledFor(level):
            # noinspection PyProtectedMember
            logger._log(level, f"run {self.command}: {message}", (data,))

    def _run_job(self):
        self.is_running = True
        self._log(logging.DEBUG, "Start")
        if self.quiet:
            stderr = DEVNULL
            stdout = DEVNULL
        else:
            stdout = PIPE
            stderr = PIPE
        try:
            process = subprocess.Popen(self.command, shell=self.shell, text=self.text, stdout=stdout, stderr=stderr)
            self.process_id = process.pid
            self._on_start()
            try:
                stdout, stderr = process.communicate(timeout=self.timeout)
            except Exception as e:
                process.kill()
                raise e
            if stdout is not None:
                if self.text:
                    self.stdout = stdout
                else:
                    self.stdout = stdout.decode()
                if self.split_stdout:
                    self.stdout = self.stdout.split("\n")[:-1]
            if stderr is not None:
                if self.text:
                    self.stderr = stderr
                else:
                    self.stderr = stderr.decode()
                if self.split_stderr:
                    self.stderr = self.stderr.split("\n")[:-1]
            self.exit_code = process.poll()
            self.is_running = False
            if self.exit_code == 0:
                self._log(logging.DEBUG, "Completed")
                self._on_success()
            else:
                self._log(logging.DEBUG, f"Error Code {self.exit_code}")
                self._on_error()
        except Exception as e:
            self.is_running = False
            self.error = e
            self._log(logging.DEBUG, f"Exception {e}")
            self._on_exception()
        return self

    def _on_start(self):
        """
            If command is starting
        """
        pass

    def _on_error(self):
        """
            If exit code of running command is not zero call this function
        """
        pass

    def _on_success(self):
        """
            If exit code of running command is zero call this function
        """
        pass

    def _on_exception(self):
        """
            If acquire any exception while running command call this function
        """
        pass

    def run(self):
        """
            Run Script
        """
        if self.__used:
            raise Exception("this command was run before. make a new instance")
        self.__used = True

        if self.background:
            job = Thread(target=self._run_job)
            job.start()
            while self.process_id is None and self.error is None:
                pass
        else:
            self._run_job()
        return self
